"""
# Linux:
cat 1.jpg 1.zip > output.jpg

# Windows:
copy/b 1.jpg+1.zip=output.jpg
"""


def encode(bytes_data: bytes, img_filename: str, img_filename_new: str):
    assert img_filename.endswith('.jpg'), "图种目前只支持 jpg 图片"
    assert img_filename_new.endswith('.jpg'), "图种目前只支持 jpg 图片"

    with open(img_filename, 'rb') as f1:
        img = f1.read()

    with open(img_filename_new, 'wb') as f2:
        f2.write(img + bytes_data)


def decode(img_filename: str) -> bytes:
    with open(img_filename, 'rb') as f:
        bytes3 = f.read()
    idx = bytes3.find(b'\xff\xd9')
    return bytes3[idx + 2:]


def file_encode(filename: str, img_filename: str, img_filename_new: str):
    with open(filename, 'rb') as f:
        encode(f.read(), img_filename, img_filename_new)


def file_decode(filename: str, img_filename: str):
    with open(filename, 'wb') as f:
        f.write(decode(img_filename))



import struct


def serialization(data: bytes) -> bytes:
    """
    功能：输入数据，输出其封装后的 bytes 类型，其封装以下内容：
    - 数据的长度（4个字节）
    - 数据本身（n个字节）
    """
    return len(data).to_bytes(length=4, byteorder='big') + data


def deserialization(serialized_data: bytes):
    """
    与 serialization 相反的操作
    """
    return serialized_data[4:4 + int.from_bytes(serialized_data[:4], byteorder="big")]


def bytes2bin_(bytes1: bytes) -> str:
    """
    把 bytes 转化为 "10110" 这种形式的二进制
    """
    return ''.join([format(i, '08b') for i in bytes1])


def bin2bytes_(bin1: str) -> bytes:
    """
    bytes2bin_ 的相反操作
    """
    return b''.join([struct.pack('>B', int(bin1[i * 8:i * 8 + 8], base=2)) for i in range(len(bin1) // 8)])


def bytes2bin(bytes1: bytes) -> list:
    """
    把 bytes 转化为 [1, 0, 1, 1, 0] 这种形式的二进制
    """
    return [int(i) for i in bytes2bin_(bytes1)]


def bin2bytes(bin1: list) -> bytes:
    """
    bytes2bin 的相反操作
    """
    return bin2bytes_(''.join([str(int(i)) for i in bin1]))



from base64 import b85encode, b85decode





def encode(bytes_data: bytes, txt_filename: str):
    with open(txt_filename, 'w') as f:
        f.write(b85encode(bytes_data).decode('utf-8'))


def decode(txt_filename: str) -> bytes:
    with open(txt_filename, 'r') as f:
        return b85decode(f.read().encode('utf-8'))


def file_encode(filename: str, txt_filename: str):
    with open(file=filename, mode='rb') as f:
        encode(bytes_data=f.read(), txt_filename=txt_filename)


def file_decode(filename: str, txt_filename: str):
    with open(file=filename, mode='wb') as f:
        f.write(decode(txt_filename=txt_filename))




# 回声音频水印
import numpy as np
from scipy.io import wavfile
from scipy.signal.windows import hann
from numpy.fft import ifft, fft


class EchoWatermark:
    def __init__(self, pwd, algo_type=3, verbose=False):
        self.pwd = pwd
        self.algo_type = algo_type
        self.verbose = verbose

        self.frame_len = 2048  # 帧长度
        self.echo_amplitude = 0.2  # 回声幅度
        self.overlap = 0.5  # 帧分析的重叠率
        self.neg_delay = 4  # negative delay, for negative echo

        # 回声参数
        # pwd[i] = 1
        self.delay11, self.delay10 = 100, 110
        # pwd[i] = 0
        self.delay01, self.delay00 = 120, 130

    def embed(self, origin_filename, wm_bits, embed_filename):
        frame_len = self.frame_len
        echo_amplitude = self.echo_amplitude
        overlap = self.overlap
        neg_delay = self.neg_delay

        delay_matrix = [[self.delay00, self.delay01], [self.delay10, self.delay11]]

        sr, ori_signal = wavfile.read(origin_filename)
        signal_len = len(ori_signal)

        # 帧的移动量
        frame_shift = int(frame_len * (1 - overlap))

        # 和相邻帧的重叠长度
        overlap_length = int(frame_len * overlap)

        # 可嵌入总比特数
        embed_nbit_ = (signal_len - overlap_length) // frame_shift

        # 重复次数
        n_repeat = embed_nbit_ // len(wm_bits)

        # 实际可嵌入的有效比特数
        len_wm_bits = len(wm_bits)
        # 实际嵌入
        embed_nbit = len_wm_bits * n_repeat

        if self.verbose:
            print(
                f"可以嵌入的总比特数为: {embed_nbit_}，水印长度为{len(wm_bits)},重复嵌入 {n_repeat} 次, 实际嵌入{embed_nbit}")

        # 扩展水印信号
        wm_repeat = np.repeat(wm_bits, n_repeat)

        # 生成密钥
        np.random.seed(self.pwd)
        secret_key = np.random.randint(2, size=int(len_wm_bits))
        secret_key_extended = np.repeat(secret_key, n_repeat)

        pointer = 0
        echoed_signal = np.zeros((frame_shift * embed_nbit))
        prev1 = np.zeros(frame_len)

        for i in range(embed_nbit):
            frame = ori_signal[pointer: (pointer + frame_len)]

            delay = delay_matrix[secret_key_extended[i]][wm_repeat[i]]

            echo_positive = np.concatenate((np.zeros(delay), frame[0:frame_len - delay]))

            echo_negative = - np.concatenate((np.zeros(delay + neg_delay),
                                              frame[0:frame_len - delay - neg_delay]))

            echo_forward = np.concatenate((frame[delay:frame_len], np.zeros(delay)))

            if self.algo_type == 1:
                echoed_frame = frame + echo_amplitude * echo_positive
            elif self.algo_type == 2:
                echoed_frame = frame + echo_amplitude * (echo_positive + echo_negative)
            else:  # algo_type == 3
                echoed_frame = frame + echo_amplitude * (echo_positive + echo_forward)

            echoed_frame = echoed_frame * hann(frame_len)
            echoed_signal[frame_shift * i: frame_shift * (i + 1)] = \
                np.concatenate((prev1[frame_shift:frame_len] +
                                echoed_frame[0:overlap_length],
                                echoed_frame[overlap_length:frame_shift]))

            prev1 = echoed_frame
            pointer += frame_shift

        echoed_signal = np.concatenate((echoed_signal, ori_signal[len(echoed_signal):])).astype(np.int16)
        # 保存为wav格式
        wavfile.write(embed_filename, sr, echoed_signal)

    def extract(self, embed_filename, len_wm_bits):
        frame_len = self.frame_len
        overlap = self.overlap
        neg_delay = self.neg_delay
        delay11, delay10, delay01, delay00 = self.delay11, self.delay10, self.delay01, self.delay00
        log_floor = 0.00001  # 取对数时的最小值

        # 打开已嵌入水印的音频文件
        _, wm_signal = wavfile.read(embed_filename)
        signal_len = len(wm_signal)

        frame_shift = int(frame_len * (1 - overlap))
        embed_nbit_ = (signal_len - int(frame_len * overlap)) // frame_shift

        # 重复次数
        n_repeat = embed_nbit_ // len_wm_bits

        # 实际可嵌入的有效比特数
        embed_nbit = len_wm_bits * n_repeat

        if self.verbose:
            print(
                f"可以嵌入的总比特数为: {embed_nbit_}，水印长度为{len_wm_bits},重复嵌入 {n_repeat} 次, 实际嵌入{embed_nbit}")

        # 加载密钥
        np.random.seed(self.pwd)
        secret_key = np.random.randint(2, size=int(len_wm_bits))
        secret_key = np.repeat(secret_key, n_repeat)

        pointer = 0
        detected_bit1 = np.zeros(embed_nbit)
        for i in range(embed_nbit):
            wmarked_frame1 = wm_signal[pointer: pointer + frame_len]
            ceps1 = ifft(
                np.log(np.square(fft(wmarked_frame1)) + log_floor)).real

            if secret_key[i] == 1:
                delay0, delay1 = delay10, delay11
            else:
                delay0, delay1 = delay00, delay01

            if self.algo_type == 1:
                if ceps1[delay1] > ceps1[delay0]:
                    detected_bit1[i] = 1
            elif self.algo_type == 2:
                if (ceps1[delay1] - ceps1[delay1 + neg_delay]) > \
                        (ceps1[delay0] - ceps1[delay0 + neg_delay]):
                    detected_bit1[i] = 1
            else:  # algo_type == 3
                if ceps1[delay1] > ceps1[delay0]:
                    detected_bit1[i] = 1

            pointer = pointer + frame_shift

        count = 0
        wm_extract = np.zeros(len_wm_bits)

        for i in range(len_wm_bits):
            # 汇总比特值（按平均值）
            ave = np.average(detected_bit1[count:count + n_repeat])
            if ave >= 0.5:
                wm_extract[i] = 1
            else:
                wm_extract[i] = 0

            count = count + n_repeat

        return wm_extract




from PIL import Image
import piexif


def encode(bytes_data: bytes, img_filename: str, img_filename_new: str):
    im = Image.open(img_filename)
    exif_data = im.info['exif']  # 二进制格式的exif

    # 借助 piexif 来修改并写入文件
    exif_dict = piexif.load(exif_data)  # 二进制格式转 dict 格式
    exif_dict['0th'][270] = bytes_data  # 写入信息
    exif_bytes = piexif.dump(exif_dict)  # dict 格式转回二进制格式
    im.save(img_filename_new, "JPEG", quality=85, exif=exif_bytes)  # 保存文件


def decode(img_filename: str) -> bytes:
    im = Image.open(img_filename)
    exif_data = im.info['exif']  # 二进制格式的exif
    exif_dict = piexif.load(exif_data)
    return exif_dict['0th'][270]



import wave
from .utils import serialization, deserialization


def encode(bytes_data: bytes, music_filename: str, music_filename_new: str):
    bytes_data = serialization(bytes_data)
    with wave.open(music_filename, "rb") as f:
        attrib = f.getparams()
        wav_data = bytearray(f.readframes(-1))

    for index in range(len(bytes_data)):
        wav_data[index * 4] = bytes_data[index]

    with wave.open(music_filename_new, "wb") as f:
        f.setparams(attrib)
        f.writeframes(wav_data)


def decode(music_filename: str) -> bytes:
    with wave.open(music_filename, 'rb') as f:
        wav_data = f.readframes(-1)

    # 得到隐藏的数据的长度
    extract_data = bytearray()
    for idx in range(4):
        extract_data += wav_data[idx * 4].to_bytes(1, byteorder='big')
    len_data = int.from_bytes(extract_data, 'big')

    extract_data = bytearray()
    for idx in range(4, 4 + len_data):
        extract_data += (wav_data[idx * 4]).to_bytes(1, byteorder='big')
    return extract_data


def file_encode(filename: str, music_filename: str, music_filename_new: str):
    with open(filename, 'rb') as f:
        encode(bytes_data=f.read(), music_filename=music_filename, music_filename_new=music_filename_new)


def file_decode(filename: str, music_filename: str):
    bytes_data = decode(music_filename=music_filename)
    with open(filename, 'wb') as f:
        f.write(bytes_data)



import struct
from .utils import serialization, deserialization
import numpy as np
from PIL import Image


def encode(bytes_data: bytes, img_filename: str, img_filename_new: str):
    data_to_write = serialization(bytes_data)
    img = np.array(Image.open(img_filename))
    height, width = img.shape[:2]

    # 二进制
    data_to_write_bin = ''.join([format(i, '08b') for i in data_to_write])
    assert len(data_to_write_bin) < height * width * 3, "要隐藏的数据太大"

    for i, binary in enumerate(data_to_write_bin):
        x, y, c = (i // 3) // width, (i // 3) % width, i % 3
        tmp = img[x, y, c]
        if binary == '0':
            img[x, y, c] = tmp & 0b11111110
        if binary == '1':
            img[x, y, c] = tmp | 0b00000001

    Image.fromarray(img).save(img_filename_new)
    return img


def decode(img_filename: str) -> bytes:
    img = np.array(Image.open(img_filename))
    height, width = img.shape[:2]

    #  前4个字节，也就32个二进制存放长度
    lst = []
    for i in range(32):
        x, y, c = (i // 3) // width, (i // 3) % width, i % 3
        lst.append(img[x, y, c] & 0b00000001)

    # 得到实际隐藏的数据长度（二进制位数）
    len_data = int(''.join(str(i) for i in lst), base=2) * 8
    lst = []
    for i in range(len_data + 32):
        x, y, c = (i // 3) // width, (i // 3) % width, i % 3
        lst.append(img[x, y, c] & 0b00000001)

    s_bin = ''.join(str(i) for i in lst)

    s_out = b''.join([struct.pack('>B', int(s_bin[i * 8:i * 8 + 8], base=2)) for i in range(len(s_bin) // 8)])

    return deserialization(s_out)


def file_encode(filename: str, img_filename: str, img_filename_new: str):
    with open(file=filename, mode='rb') as f:
        encode(bytes_data=f.read(), img_filename=img_filename, img_filename_new=img_filename_new)


def file_decode(filename: str, img_filename: str):
    with open(file=filename, mode='wb') as f:
        f.write(decode(img_filename=img_filename))



import math
import numpy as np
from PIL import Image
from .utils import serialization, deserialization


def encode(bytes_data: bytes, img_filename: str):
    data_to_write = serialization(bytes_data)

    len_data = len(data_to_write)
    width = math.ceil((len_data / 3) ** 0.5)
    data_img = np.zeros((width, width, 3), dtype=np.uint8)
    for i, u8 in enumerate(data_to_write):
        data_img[(i // 3) // width, (i // 3) % width, i % 3] = u8

    Image.fromarray(data_img).save(img_filename)


def decode(img_filename: str) -> bytes:
    img = Image.open(img_filename)

    width, height = img.size
    lst = []

    for y in range(height):
        for x in range(width):
            rgb = img.getpixel((x, y))
            lst.extend(rgb)

    return deserialization(bytes(lst))


def file_encode(filename: str, img_filename: str):
    with open(file=filename, mode='rb') as f:
        encode(bytes_data=f.read(), img_filename=img_filename)


def file_decode(filename: str, img_filename: str):
    with open(file=filename, mode='wb') as f:
        f.write(decode(img_filename=img_filename))



import numpy as np


def get_error_rate(wm_extract, wm_bits):
    len_wm_bits = len(wm_bits)
    error_num = np.sum(np.abs(wm_extract - wm_bits))
    error_rate = error_num / len_wm_bits
    print(f'bit error rate = {error_rate:.2%} ({error_num} / {len_wm_bits})')
    return error_rate


def get_snr(ori_signal, wm_signal):
    snr = 10 * np.log10(
        np.sum(np.square(ori_signal.astype(np.float32)))
        / np.sum(np.square(ori_signal.astype(np.float32)
                           - wm_signal.astype(np.float32))))
    print(f'SNR = {snr:.2f} dB')
    return snr



import numpy
import numpy as np
from PIL import Image


def encode(img: numpy.ndarray, watermark: numpy.ndarray, n: int):
    height, width = img.shape[:2]
    height0, width0 = watermark.shape

    a1 = np.uint8(1 << n)
    a2 = ~(np.uint8(a1))

    for i in range(height):
        for j in range(width):
            if watermark[i % height0, j % width0]:
                for channel in range(3):
                    img[i, j, channel] |= a1
            else:
                for channel in range(3):
                    img[i, j, channel] &= a2


def decode(img, n):
    height, width = img.shape[:2]
    a1 = np.uint8(1 << n)
    watermark_extract = np.zeros(shape=(height, width), dtype=np.uint8)

    for i in range(height):
        for j in range(width):
            tmp = sum(img[i, j, channel] & a1 for channel in range(3))
            watermark_extract[i, j] = int(tmp * 85)  # 85 = 255 / 3

    return watermark_extract


def file_encode(img_filename: str, watermark_filename: str, img_filename_new: str, n: int = 0):
    # n：0～3，越大水印强度越高
    watermark = np.array(Image.open(watermark_filename).convert('1'))
    img = np.array(Image.open(img_filename))
    encode(img, watermark, n)
    Image.fromarray(img).save(img_filename_new)


def file_decode(img_filename: str, wm_extract: str, n: int = 0):
    img = np.array(Image.open(img_filename))
    watermark_extract = decode(img, n)
    Image.fromarray(watermark_extract).save(wm_extract)




import numpy as np
import wave
from scipy import fftpack

SAMPLE_RATE = 44100  # 音频采样率
SAMPLE_TIME = 0.05  # 采样时间
SAMPLE_NUM = int(SAMPLE_RATE * SAMPLE_TIME)  # 采样点数量

FREQUENCY_LIST = [i * 100 + 40 for i in range(16)]  # 选择 16 个音（单位Hz）表示一个四进制
QUA_LIST = '0123456789abcdef'
FREQUENCY_DICT = dict(zip(QUA_LIST, FREQUENCY_LIST))

X = np.arange(0, SAMPLE_TIME, 1 / SAMPLE_RATE)  # 用于生成波形


def encode1(qua):
    # 一位四进制数 -> 波形
    return np.sin(2 * np.pi * FREQUENCY_DICT[qua] * X)


def encode_all(bytes_data: bytes):
    return np.concatenate([encode1(c) for c in bytes_data.hex()])


def encode(bytes_data: bytes, wav_filename: str):
    wav_data = encode_all(bytes_data)
    with wave.open(wav_filename, mode='wb') as f:
        f.setparams((1, 2, SAMPLE_RATE, len(wav_data), "NONE", "NOT COMPRESSED"))  # 单通道
        f.writeframes((wav_data * 32768).astype(np.int16))  # wav_data原幅值为[-1,1]，乘32768可以提高音量


def fft(data):
    N = len(data)
    fft_data = fftpack.fft(data)
    abs_fft = np.abs(fft_data)
    abs_fft = abs_fft / (N / 2)  # fft后弦波分量的振幅放大了N/2倍，除一下以归一化
    half_fft = abs_fft[range(N // 2)]  # 取单边频谱
    return half_fft


def decode1(wav_slice):
    fft_ret = fft(wav_slice)
    for idx, freq in enumerate(FREQUENCY_LIST):
        if np.max(fft_ret[int(freq * SAMPLE_TIME) - 2: int(freq * SAMPLE_TIME) + 2]) > 0.8:
            return QUA_LIST[idx]


def decode(wav_filename: str):
    with wave.open(wav_filename, 'rb') as f:
        wav_data = np.frombuffer(f.readframes(-1), dtype=np.int16) / 32768

    len_data = len(wav_data) // SAMPLE_NUM  # 音频片段个数
    hex_data = ''.join(decode1(wav_data[i * SAMPLE_NUM: (i + 1) * SAMPLE_NUM]) for i in range(len_data))
    return bytes.fromhex(hex_data)


def file_encode(filename: str, wav_filename: str):
    with open(file=filename, mode='rb') as f:
        encode(bytes_data=f.read(), wav_filename=wav_filename)


def file_decode(filename: str, wav_filename: str):
    with open(file=filename, mode='wb') as f:
        f.write(decode(wav_filename=wav_filename))



'''
原理如下：
- 目标是使得一张PNG图片（带透明涂层），在白色背景下显示图A，在黑色背景下显示图B
- 这样的要求下，可以列出以下方程

aP +(1-a)白 = A
aP + (1-a)黑 = B

已知 白色 = 255，黑色 = 0，代入方程，得到：
a = 1- (A-B)
P = B/a

因为 a 的取值范围为 [0, 1]，因此需要 B 图较暗，A图较亮
'''

import cv2
import numpy as np


# 这个只能输出黑白图
def mirage_tank2(img_filename1: str, img_filename2: str, output_img_filename: str):
    A = cv2.imread(img_filename1, flags=cv2.IMREAD_GRAYSCALE) / 255
    B = cv2.imread(img_filename2, flags=cv2.IMREAD_GRAYSCALE) / 255

    # 保证两个图片的大小一致
    B = cv2.resize(B, A.shape[::-1])
    a = 1 - (A - B)
    a = np.clip(a, 0, 1)
    P = B / a
    P = np.clip(P, 0, 1)

    # 创建一个带有透明度通道的图像
    image_with_alpha = np.zeros((A.shape[0], A.shape[1], 4), dtype=np.uint8)
    image_with_alpha[:, :, 0] = P * 255
    image_with_alpha[:, :, 1] = P * 255
    image_with_alpha[:, :, 2] = P * 255
    image_with_alpha[:, :, 3] = a * 255

    # 保存带有透明度通道的图像
    cv2.imwrite(output_img_filename, image_with_alpha)


def mirage_tank(img_filename1: str, img_filename2: str, output_img_filename: str, a=0.5, b=20):
    A = cv2.imread(img_filename1)
    B = cv2.imread(img_filename2)

    height, width, _ = A.shape
    # 保证两个图片的大小一致
    B = cv2.resize(B, (width, height))

    # 计算 alpha 通道
    A_gray = cv2.cvtColor(A, cv2.COLOR_BGR2GRAY)
    B_gray = cv2.cvtColor(B, cv2.COLOR_BGR2GRAY)
    B_gray = a * B_gray + b  # 适当调暗图片B
    alpha = 255 - A_gray.astype('int') + B_gray.astype('int')
    alpha = np.clip(alpha, 1, 255).reshape(height, width, 1)

    # 使用 A 或者 B，决定了彩图来自哪个图
    # P = B / (a / 255)
    P = (A - (255 - alpha)) / (alpha / 255)

    alpha = alpha.astype('u8')
    P = np.clip(P, 0, 255)

    # 带有透明度通道的图像
    image_with_alpha = np.concatenate([P, alpha], axis=2)

    cv2.imwrite(output_img_filename, image_with_alpha)



from .utils import serialization, deserialization
import struct


def encode(bytes_data: bytes, text: str):
    data_to_write = serialization(bytes_data)

    data_to_write_bin = ''.join([format(i, '08b') for i in data_to_write])
    assert len(data_to_write_bin) > len(bytes_data) + 1, "要隐藏的数据太大"

    text_new = ''

    for idx, word in enumerate(text):
        text_new += word
        if idx < len(data_to_write_bin) and data_to_write_bin[idx] == '1':
            text_new += chr(127)
    return text_new


def decode(text_new: str) -> bytes:
    bytes_data_bin_extract = ''
    idx = 1
    while idx < len(text_new):
        if text_new[idx] != chr(127):
            idx += 1
            bytes_data_bin_extract += '0'
        else:
            idx += 2
            bytes_data_bin_extract += '1'

    s_bin = bytes_data_bin_extract
    s_out = b''.join([struct.pack('>B', int(s_bin[i * 8:i * 8 + 8], base=2)) for i in range(len(s_bin) // 8)])

    return deserialization(s_out)


def file_encode(filename: str, text_filename: str, text_filename_new: str):
    with open(file=filename, mode='rb') as f1 \
            , open(file=text_filename, mode='r') as f2 \
            , open(file=text_filename_new, mode='w') as f3:
        text_new = encode(bytes_data=f1.read(), text=f2.read())
        f3.write(text_new)


def file_decode(filename: str, text_filename: str):
    with open(file=filename, mode='wb') as f, open(file=text_filename, mode='r') as f2:
        f.write(decode(text_new=f2.read()))
